"""
pansat.time
===========

This module provides functionality for dealing with times.
"""
from datetime import datetime
from dataclasses import dataclass
import pandas as pd
import numpy as np


def to_datetime(time):
    """
    Try to convert a given time to a datetime object.
    """
    if isinstance(time, datetime):
        return time
    try:
        return pd.to_datetime(time).to_pydatetime()
    except ValueError:
        raise ValueError("Could not convert '%s' to datetime object.")


def to_datetime64(time):
    """
    Try to convert a given time to a numpy datetime64 object.
    """
    if isinstance(time, np.ndarray) and time.dtype == np.datetime64:
        return time
    try:
        return pd.to_datetime(time).to_datetime64()
    except ValueError:
        raise ValueError("Could not convert '%s' to datetime object.")


@dataclass
class TimeRange:
    """
    A time range defining the temporal extent of a dataset.

    The time range is represented by the 'start' and 'end' attributes
    of the class representing the start and end of the time range covered
    by a given data filee.

    If the temporal extent of a data file cannot be deduced from the
    filename alone or it is not known, the 'end' attribute can be 'None'.
    """

    start: np.datetime64
    end: np.datetime64 = None



"""
==================================
pansat.products.satellite.meteosat
==================================

This module provides product classes and object for satellite products
derived from the Meteosat Second Generation (MSG) satellites.
"""
from pathlib import Path
import re
from zipfile import ZipFile
from datetime import datetime

import pansat.download.providers as providers
from pansat.products.product import Product
from pansat.exceptions import NoAvailableProvider


def _extract_file(filename):
    """
    Extracts the data file from the .zip archive downloaded from the
    provider and deletes the original archive.
    """
    path = Path(filename)
    data = path.stem + ".nat"
    with ZipFile(path) as archive:
        archive.extract(data, path=path.parent)
    path.unlink()
    return path.parent / data


class MSGSeviriL1BProduct(Product):
    """
    Base class for Meteosat Second Generation (MSG) SEVIRI L1B products.
    """

    def __init__(self, location=None):
        """
        Create MSG Seviri L1B product.

        Args:
            location: None for the 0-degree position of MSG and "IO" for the
                 the position over the Indian Ocean.

        """
        self.name = "MSG_Seviri"

        if location is not None:
            if location == "IO":
                self.name = "MSG_Seviri_IO"
            else:
                raise ValueError(
                    "'location' kwarg of MSGSeviriProduct should be None for "
                    " the 0-degree position or 'IO' for the Indian Ocean "
                    "location."
                )
        self.filename_regex = re.compile(
            "MSG\d-SEVI-MSG15-0100-NA-(\d{14})\.\d*Z-NA.nat"
        )

    @property
    def default_destination(self):
        return Path("MSG")

    def __str__(self):
        return self.name

    def filename_to_date(self, filename):
        match = self.filename_regex.match(Path(filename).name)
        if match is None:
            raise ValueError(
                f"Given filename '{filename}' does not match the expected "
                f"filename format of MSG Seviri L1B files."
            )
        time = datetime.strptime(match.group(1), "%Y%m%d%H%M%S")
        return time

    def _get_provider(self):
        """Find a provider that provides the product."""
        available_providers = [
            p
            for p in providers.ALL_PROVIDERS
            if str(self) in p.get_available_products()
        ]
        if not available_providers:
            raise NoAvailableProvider(
                f"Could not find a provider for the" f" product {self.name}."
            )
        return available_providers[0]

    def download(self, start_time, end_time, destination=None, provider=None):
        """
        Download data product for given time range.

        Args:
            start_time(``datetime``): ``datetime`` object defining the start
                 date of the time range.
            end_time(``datetime``): ``datetime`` object defining the end date
                 of the of the time range.
            destination(``str`` or ``pathlib.Path``): The destination where to
                 store the output data.
        """

        if not provider:
            provider = self._get_provider()

        if not destination:
            destination = self.default_destination
        else:
            destination = Path(destination)
        destination.mkdir(parents=True, exist_ok=True)
        provider = provider(self)

        files = provider.download(start_time, end_time, destination)
        return [_extract_file(f) for f in files]


l1b_msg_seviri = MSGSeviriL1BProduct()
l1b_msg_seviri_io = MSGSeviriL1BProduct(location="IO")



"""
pansat.products.ground_based
============================

This module defines products provided by Cloudnet. The official
data portal for CloudNet is available
`here <https://cloudnet.fmi.fi/>`_.
"""
from datetime import datetime
from pathlib import Path
import re

from pansat.exceptions import NoAvailableProvider
from pansat.products.product import Product
from pansat.download import providers


class CloudnetProduct(Product):
    """
    This class represents all Cloudnet products. Since Cloudnet data
    is derived from specific stations a product can have an associated
    location in which case only data of the product collected at the
    given location is represetned by the product.
    """

    def __init__(self, product_name, description, location=None):
        """
        Args:
            product_name: The name of the product.
            description: A string describing the product.
            location: An optional string specifying the a specific
                 Cloudnet location.
        """
        self.product_name = product_name
        self._description = description
        self.location = location

        if location is not None:
            self.filename_regexp = re.compile(rf"(\d{{8}})_{location}_[-\w\d]*.nc")
        else:
            self.filename_regexp = re.compile(rf"(\d{{8}})_([\w-]*)_[-\w\d]*.nc")

    @property
    def description(self):
        return self._description

    def matches(self, filename):
        """
        Determines whether a given filename matches the pattern used for
        the product.

        Args:
            filename(``str``): The filename

        Return:
            True if the filename matches the product, False otherwise.
        """
        return self.filename_regexp.match(filename)

    def filename_to_date(self, filename):
        """
        Extract timestamp from filename.

        Args:
            filename(``str``): Filename of a Cloudnet product.

        Returns:
            ``datetime`` object representing the timestamp of the
            filename.
        """
        parts = filename.split("/")
        if len(parts) > 1:
            filename = parts[-1]
        path = Path(filename)
        match = self.filename_regexp.match(path.name)

        date_string = match.group(1)
        date = datetime.strptime(date_string, "%Y%m%d")
        return date

    def _get_provider(self):
        """Find a provider that provides the product."""
        available_providers = [
            p
            for p in providers.ALL_PROVIDERS
            if str(self) in p.get_available_products()
        ]
        if not available_providers:
            raise NoAvailableProvider(
                f"Could not find a provider for the" f" product {self}."
            )
        return available_providers[0]

    @property
    def default_destination(self):
        """
        The default location for Cloudnet products is cloudnet/<product_name>
        """
        return Path("cloudnet") / Path(self.product_name)

    def __str__(self):
        s = f"ground_based::Cloudnet::{self.product_name}"
        return s

    def download(self, start_time, end_time, destination=None, provider=None):
        """
        Download data product for given time range.

        Args:
            start_time(``datetime``): ``datetime`` object defining the start
                 date of the time range.
            end_time(``datetime``): ``datetime`` object defining the end date
                 of the of the time range.
            destination(``str`` or ``pathlib.Path``): The destination where to
                 store the output data.
        """
        if not provider:
            provider = self._get_provider()

        if not destination:
            destination = self.default_destination
        else:
            destination = Path(destination)
        destination.mkdir(parents=True, exist_ok=True)
        provider = provider(self)

        return provider.download(start_time, end_time, destination)

    def open(self, filename):
        """
        Open file as xarray dataset.

        Args:
            filename(``pathlib.Path`` or ``str``): The Cloudnet file to open.
        """
        return xr.load_dataset(filename)


l2_iwc = CloudnetProduct("iwc", "IWC calculated from Z-T method.")
l1_radar = CloudnetProduct("radar", "L1b radar data.")



"""
pansat.products.satellite.gridsat
=================================

This module defines the NOAA GridSat products.
"""
from datetime import datetime
from pathlib import Path

from pansat.download import providers
from pansat.exceptions import NoAvailableProvider


class GridsatProduct:
    """
    Class for NOAA GridSat GOES and CONUS products.
    """

    def __init__(self, variant):
        """
        Args:
            variant: The variant of the GridSat product: 'conus' or 'goes'.
        """
        self.variant = variant

    def filename_to_date(self, filename):
        """
        Args:
            filename: The name of GridSat file.
        Return:
            ``datetime.datetime`` object of the corresponding
            time.
        """
        parts = filename.split(".")
        year, month, day, hour_min = parts[2:6]
        return datetime(
            int(year), int(month), int(day), int(hour_min[:2]), int(hour_min[2:])
        )

    @property
    def name(self):
        """The product name."""
        return f"gridsat_{self.variant}"

    @property
    def default_destination(self):
        """Default destination for downloads."""
        return "grid_sat"

    def open(self, filename):
        """Open given file as ``xarray.Dataset``."""
        return xr.open_dataset(self.filename)

    def _get_provider(self):
        """Find a provider that provides the product."""
        available_providers = [
            p
            for p in providers.ALL_PROVIDERS
            if self.name in p.get_available_products()
        ]
        if not available_providers:
            raise NoAvailableProvider(
                f"Could not find a provider for the" f" product {self.name}."
            )
        return available_providers[0]

    def download(self, start_time, end_time, destination=None, provider=None):
        """
        Download data product for given time range.

        Args:
            start_time(``datetime``): ``datetime`` object defining the start
                 date of the time range.
            end_time(``datetime``): ``datetime`` object defining the end date
                 of the of the time range.
            destination(``str`` or ``pathlib.Path``): The destination where to
                 store the output data.
        """

        if not provider:
            provider = self._get_provider()

        if not destination:
            destination = Path(self.default_destination)
        else:
            destination = Path(destination)
        destination.mkdir(parents=True, exist_ok=True)
        provider = provider(self)

        files = provider.download(start_time, end_time, destination)
        return files


class GridsatB1(GridsatProduct):
    """
    Specialized class for the GridSat CDR.
    """

    def __init__(self):
        """Create product."""
        super().__init__("b1")

    def filename_to_date(self, filename):
        """
        Args:
            filename: The name of GridSat file.
        Return:
            ``datetime.datetime`` object of the corresponding
            time.
        """
        parts = filename.split(".")
        year, month, day, hour = parts[1:5]
        return datetime(int(year), int(month), int(day), int(hour))


gridsat_goes = GridsatProduct("goes")
gridsat_conus = GridsatProduct(("conus"))
gridsat_b1 = GridsatB1()



"""
pansat.download.providers.cloudsat_dpc
======================================

This module defines a provider for the CloudSat DPC SFTP server. To use this
 provider you need to setup your CloudSat DPC account
at https://www.cloudsat.cira.colostate.edu/data-products/2c-ice.

Setup
`````

The CloudSatDPC provider provides access to the SFTP servers of the
 CloudSat DPC. To use it, you need to add a public SSH key to your
account at the DPC website. To connect to the SFTP server,
``pansat`` needs access to the corresponding private key. Therefore,
you will have to add
 1. your SFTP username and
 2. the SSH private key corresponding to the public key you
    have uploaded to the DPC website
as username and password in your pansat identities file.

The private key can be provided either as a path pointing to a
 local file, or by directly storing the base-64 encoded key
as the password.

.. warning::
   Make sure to create a separate SSH key pair for this purpose.
   Do not use the same keys that you are using for other
   purposes as you cannot expect them to be save when added to
   pansat.
"""
from datetime import datetime
from io import StringIO
import paramiko
from pathlib import Path

from pansat.download.accounts import get_identity
from pansat.download.providers.discrete_provider import DiscreteProvider


PRODUCTS = {
    "CloudSat_2C-ICE": "2C-ICE.P1_R05",
    "CloudSat_2B-CLDCLASS": "2B-CLDCLASS.P1_R05",
    "CloudSat_2B-CLDCLASS-LIDAR": "2B-CLDCLASS-LIDAR.P1_R05",
    "CloudSat_2C-RAIN PROFILE": "2C-RAIN-PROFILE.P1_R05",
}

######################################################################
# SFTP Connection
######################################################################


class SFTPConnection:
    """
    Helper class that manages the lifetime of a paramiko SFTP connection.
    """

    def __init__(self, host, provider):
        """
        Open connection to host. Credentials are obtained from the pansat
        identities database.
        """
        self.host = host
        self.provider = provider

        self.transport = None
        self.sftp = None
        self._connect()

    def _connect(self):
        user_name, key = get_identity(self.provider)

        if not Path(key).exists():
            key_buffer = StringIO()
            key_buffer.write("-----BEGIN OPENSSH PRIVATE KEY-----\n")
            key_buffer.write(key + "\n")
            key_buffer.write("-----END OPENSSH PRIVATE KEY-----\n")
        else:
            with open(key, "r") as key_file:
                key_buffer = StringIO(key_file.read())
        key_buffer.seek(0)
        key = paramiko.RSAKey.from_private_key(key_buffer)
        self.transport = paramiko.Transport(self.host)
        self.transport.connect(username=user_name, pkey=key)
        self.sftp = paramiko.SFTPClient.from_transport(self.transport)

    def list_files(self, path):
        """
        List files in given remote directory.

        Args:
            path: The path to the directory for which to list the files.

        Return:
            List containing the files in the given directory.
        """
        return self.sftp.listdir(path)

    def download(self, path, destination):
        """
        Download file to destination.

        Args:
            path: The remote path of the file to download.
            destination: Path to the file to which to write
                the results.
        """
        self.sftp.getfo(path, open(destination, "wb"))

    def ensure_connection(self):
        """
        Ensure that SSH connection is still alive.
        """
        try:
            self.transport.send_ignore()
        except EOFError:
            self._connect()

    def __del__(self):
        """Close connection."""
        if self.sftp is not None:
            self.sftp.close()
            self.sftp = None
        if self.transport is not None:
            self.transport.close()
            self.transport = None


######################################################################
# SFTP Connection
######################################################################


class CloudSatDPCProvider(DiscreteProvider):
    """
    Data provider class for the CloudSat DPC SFTP server.
    """

    def __init__(self, product):
        """
        Args:
            product: The pansat product to download.
        """
        super().__init__(product)
        self.product = product
        self._connection = None

    @property
    def connection(self):
        """SFTP connection object to the data server."""
        if self._connection is None:
            self._connection = SFTPConnection(
                "www.cloudsat.cira.colostate.edu", "CloudSatDPC"
            )
        return self._connection

    @connection.setter
    def connection(self, con):
        self._connection = con

    @staticmethod
    def get_available_products():
        return list(PRODUCTS.keys())

    def get_files_by_day(self, year, day):
        """
        Get the available for a given day of a year.

        Args:
            year: The year
            day: The Julian day

        Return:
            List of files available for the given day.
        """
        self.connection = SFTPConnection(
            "www.cloudsat.cira.colostate.edu", "CloudSatDPC"
        )
        try:
            directory = f"/Data/{PRODUCTS[self.product.name]}/{year:04}/{day:03}"
            self.connection.ensure_connection()
            return self.connection.list_files(directory)
        except FileNotFoundError as e:
            return []

    def download_file(self, filename, destination):
        """
        Download a given file and write the results to the given destination.

        Args:
            filename: The filename of the file to download.
            destination: The destination to which to write the downloaded file.
        """
        date = self.product.filename_to_date(filename)

        year = date.year
        day = (date - datetime(date.year, 1, 1)).days + 1
        path = f"/Data/{PRODUCTS[self.product.name]}/" f"{year:04}/{day:03}/{filename}"
        self.connection.ensure_connection()
        self.connection.download(path, destination)



import numpy as np
from shapely.geometry import Polygon, MultiPolygon
from shapely.validation import make_valid
from shapely import unary_union


def parse_point(xml_point):
    """
    Parse point from XML file.
    """
    lon = float(xml_point[0].text)
    lat = float(xml_point[1].text)
    return [lon, lat]


def parse_polygon(xml_polygon):
    """
    Parse polygon from XML file.
    """
    boundary = xml_polygon[0]
    points = np.array(list(map(parse_point, boundary.getchildren())))
    dlons = points[:, 0] - points[0, 0]
    indices = np.where(np.abs(dlons) > 180)[0]
    for ind in indices:
        if points[ind, 0] > 0:
            points[ind, 0] -= 360
        else:
            points[ind, 0] += 360

    return make_valid(Polygon(points))


def parse_swath(meta_data):
    """
    Parse shapes describing a satellite swath from metadata.

    Args:
        meta_data: XML tree containing the meta data.

    Return:
        List of polygons.
    """
    sdc = meta_data.find("SpatialDomainContainer")
    hsdc = sdc.find("HorizontalSpatialDomainContainer")

    polygons = list(map(parse_polygon, hsdc.getchildren()))
    for ind in range(len(polygons)):
        poly = polygons[ind]
        points = np.array(poly.convex_hull.exterior.coords)
        if any(points[:, 1] > 70):
            poly_2 = Polygon([[-180, 75], [180, 75], [180, 90], [-180, 90]])
            polygons[ind] = poly.union(poly_2)
        if any(points[:, 1] < -70):
            poly_2 = Polygon([[-180, -75], [180, -75], [180, -90], [-180, -90]])
            polygons[ind] = poly.union(poly_2)

    return unary_union(polygons)


def reshape_polygon(multi_polygon):
    to_centroid = lambda x: np.array(x.centroid())
    centroids = map()
    polys = multi_polygon.geoms



"""
pansat.products.satellite.mhs
=============================

This module defines the MHS product group which comprises observations from
any of the MHS sensors.
"""
import re
from datetime import datetime
from pathlib import Path

import numpy as np

import pansat.download.providers as providers
from pansat.products.product import Product
from pansat.products.product_description import ProductDescription
from pansat.exceptions import NoAvailableProvider


class MHSProduct(Product):
    """
    Base class representing MHS products.
    """

    def __init__(self):
        self.filename_regexp = re.compile(
            "MHSx_xxx_1B_M01_(\d{8})Z_(\d{8})Z_N_O_\d{8}Z.nat"
        )
        self.name = "MHS_L1B"

    @property
    def variables(self):
        return []

    @property
    def description(self):
        return self._description

    def matches(self, filename):
        """
        Determines whether a given filename matches the pattern used for
        the product.

        Args:
            filename(``str``): The filename

        Return:
            True if the filename matches the product, False otherwise.
        """
        return self.filename_regexp.match(filename)

    def filename_to_date(self, filename):
        """
        Extract timestamp from filename.

        Args:
            filename(``str``): Filename of a GPM product.

        Returns:
            ``datetime`` object representing the timestamp of the
            filename.
        """
        path = Path(filename)
        match = self.filename_regexp.match(path.name)
        date_string = match.group(2) + match.group(3)
        date = datetime.strptime(date_string, "%Y%m%d%H%M%S")
        return date

    def _get_provider(self):
        """Find a provider that provides the product."""
        available_providers = [
            p
            for p in providers.ALL_PROVIDERS
            if str(self) in p.get_available_products()
        ]
        if not available_providers:
            raise NoAvailableProvider(
                f"Could not find a provider for the" f" product {str(self)}."
            )
        return available_providers[0]

    @property
    def default_destination(self):
        """
        The default destination for GPM product is
        ``GPM/<product_name>``>
        """
        return Path("MHS") / Path(self.name)

    def __str__(self):
        return "MHS_L1B"

    def download(self, start_time, end_time, destination=None, provider=None):
        """
        Download data product for given time range.

        Args:
            start_time(``datetime``): ``datetime`` object defining the start
                 date of the time range.
            end_time(``datetime``): ``datetime`` object defining the end date
                 of the of the time range.
            destination(``str`` or ``pathlib.Path``): The destination where to
                 store the output data.
        """

        if not provider:
            provider = self._get_provider()

        if not destination:
            destination = self.default_destination
        else:
            destination = Path(destination)
        destination.mkdir(parents=True, exist_ok=True)
        provider = provider(self)

        return provider.download(start_time, end_time, destination)

    def open(self, filename):
        """
        Open file as xarray dataset.

        Args:
            filename(``pathlib.Path`` or ``str``): The GPM file to open.
        """
        from pansat.formats.hdf5 import HDF5File

        file_handle = HDF5File(filename, "r")
        return self.description.to_xarray_dataset(file_handle, globals())


l1b_mhs = MHSProduct()



"""
pansat.file_record
==================

Defines a file record class that contains information about a local or remote
file.
"""
from dataclasses import dataclass
from pathlib import Path


@dataclass
class FileRecord:
    """
    A file record represents a data file of a given product that is either
    available locally or remotely.
    """

    product: "pansat.products.Product"
    filename: str
    local_path: Path = None
    remote_path: str = None

    @staticmethod
    def from_remote(
        product: "pansat.products.Product", remote_path: str, filename: str
    ):
        """
        Create file record from a remote file.

        Args:
            product: A pansat product representing the data product.
            remote_path: String describing the remote location of the
                data file.
            filename: The filename of the data file.

        Return:
            A 'FileRecord' object representing the file.
        """
        rec = FileRecord(product, None)
        rec.remote_path = remote_path
        rec.filename = filename
        return rec

    def __init__(self, product, local_path):
        """
        Create file record from product and local file path.

        Args:
            product: A pansat product representing the data product.
            local_path: The local path of the file.
        """
        self.product = product
        local_path = Path(local_path)
        self.filename = local_path.name
        self.local_path = local_path
        self.remote_path = None



"""
pansat.download.providers.noaa_ncei
===================================

This module defines a provider for the NOAA NCEI data server containing
the GridSat brightness temperature datasets available at
 https://www.ncei.noaa.gov/data/
"""
from datetime import datetime, timedelta
import re

import requests
from pansat.download.providers.discrete_provider import DiscreteProvider


BASE_URL = "https://www.ncei.noaa.gov/data"


NCEI_PRODUCTS = {
    "gridsat_goes": "gridsat-goes/access/goes",
    "gridsat_conus": "gridsat-goes/access/conus",
    "gridsat_b1": "geostationary-ir-channel-brightness-temperature-gridsat-b1/access",
}


class NOAANCEIProvider(DiscreteProvider):
    """
       Data provider for GridSat GOES datasets  available at
    https://www.ncei.noaa.gov/data/.
    """

    def __init__(self, product):
        """
        Instantiate provider for given product.

        Args:
            product: Product instance provided by the provider.
        """
        super().__init__(product)
        self.product = product
        self.cache = {}

    @classmethod
    def get_available_products(cls):
        """
        Return the names of products available from this data provider.

        Return:
            A list of strings containing the names of the products that can
            be downloaded from this data provider.
        """
        return NCEI_PRODUCTS.keys()

    def get_files_by_month(self, year, month):
        """
        Get files available in a given month.

        Args:
            year: The year as ``int``.
            month: The month of the year as ``int``.

        Return:
            A list of the available files.
        """
        url = f"{BASE_URL}/{NCEI_PRODUCTS[self.product.name]}/{year:04}/{month:02}"
        response = requests.get(url)
        # Some products aren't split up by month. So try to get files by year
        # instead.
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError:
            url = f"{BASE_URL}/{NCEI_PRODUCTS[self.product.name]}/{year:04}/"
            response = requests.get(url)

        pattern = re.compile(r'<a href="([^"]*\.nc)">')
        return pattern.findall(response.text)

    def get_files_by_day(self, year, day):
        """
        Get files available in a given day.

        Args:
            year: The year as ``int``.
            day: The day of the year as ``int``.

        Return:
            A list of the available files.
        """
        date = datetime(year, 1, 1) + timedelta(days=day - 1)
        month = date.month
        day_of_month = date.day

        files = self.cache.get((year, month))
        if files is None:
            files = self.get_files_by_month(year, month)
            self.cache[(year, month)] = files

        dates = map(self.product.filename_to_date, files)
        files = [
            name
            for name, date in zip(files, dates)
            if date.day == day_of_month and date.month == month
        ]
        return files

    def download_file(self, filename, destination):
        """
        Download the file to a given destination.

        Args:
            filename: Name of the file to download.
            destination: The destination to which to write the
                results.
        """
        date = self.product.filename_to_date(filename)
        year = date.year
        month = date.month

        url = f"{BASE_URL}/{NCEI_PRODUCTS[self.product.name]}/{year:04}/{month:02}/{filename}"

        response = requests.get(url)
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError:
            url = f"{BASE_URL}/{NCEI_PRODUCTS[self.product.name]}/{year:04}/{filename}"
            response = requests.get(url)
            response.raise_for_status()

        with open(destination, "wb") as output:
            for chunk in response:
                output.write(chunk)



"""
pansat.products
===============

The ``products`` module provides functionality for handling supported data products.
"""
from abc import ABC, abstractmethod
from pathlib import Path

import xarray as xr

from pansat.download.providers.copernicus import COPERNICUS_PRODUCTS
from pansat.download.providers.noaa import NOAA_PRODUCTS
from pansat.download.providers.icare import ICARE_PRODUCTS
from pansat.download.providers.ges_disc import GPM_PRODUCTS
from pansat.file_record import FileRecord
from pansat.time import TimeRange
from pansat import geometry

ALL_PRODUCTS = [
    *COPERNICUS_PRODUCTS,
    *NOAA_PRODUCTS,
    *list(ICARE_PRODUCTS.keys()),
    *list(GPM_PRODUCTS.keys()),
]


class Geometry:
    """A dummy class."""

    pass


class Product(ABC):
    """
    Generic interface for datasets managed by pansat.

    This class defines the essential functionality that a product must
    implement to be used inside the pansat framework.
    """

    @abstractmethod
    def matches(self, path: Path) -> bool:
        """
        Determine wheter a given file belongs to this product.

        Args:
            path: A path object pointing to a local file.

        Return:
            'True' if the file belongs to this product. 'False' otherwise.
        """

    @abstractmethod
    def get_temporal_coverage(self, rec: FileRecord) -> TimeRange:
        """
        Determine the temporal coverage of a data file.

        This function should try to deduce the time coverage of a data file
        from its filename.

        If it cannot be avoided, a local file may be opened to determine the
        time range. However, this should only be performed if the reading of
        the time range is efficient.

        Args:
            rec: A 'FileRecord' object representing the file from which to
                deduce the temporal coverage.

        Return:
            A 'TimeRange' object representing the time range covered by the
            data file.
        """

    @abstractmethod
    def get_spatial_coverage(self, rec: FileRecord) -> Geometry:
        """
        Determine the spatial coverage of a data file.

        Args:
            rec: A file record representing the file of which to determine
                the spatial extent.

        Return:
            A 'Geometry' object representing the spatial that the given
            datafile covers.
        """
        pass

    @abstractmethod
    def open(self, rec: FileRecord) -> xr.Dataset:
        """
        This function should read a given data file into an xarray.Dataset.

        Args:
            rec: A file record representing the file to be opened.

        Return:
            An 'xarray.Dataset' that contains the data of the provided file.
        """
        pass


class NetcdfProduct(ABC):
    """
    A generic product interface for data in NetCDF format with certain standard
    names.
    """

    def __init__(self, variable_names=None):
        super().__init__()
        self.variable_names = variable_names

    def _get_variable_name(self, name):
        if self.variable_names is None:
            return name
        return self.variable_names.get(name, name)

    def matches(self, path: Path) -> bool:
        """
        Determine wheter a given file belongs to this product.

        Args:
            path: A path object pointing to a local file.

        Return:
            'True' if the file belongs to this product. 'False' otherwise.
        """
        return True

    def get_temporal_coverage(self, rec: FileRecord) -> TimeRange:
        """
        Determine the temporal coverage of a data file.

        This function should try to deduce the time coverage of a data file
        from its filename.

        If it cannot be avoided, a local file may be opened to determine the
        time range. However, this should only be performed if the reading of
        the time range is efficient.

        Args:
            rec: A 'FileRecord' object representing the file from which to
                deduce the temporal coverage.

        Return:
            A 'TimeRange' object representing the time range covered by the
            data file.
        """
        if rec.local_path is not None and rec.local_path.exists():
            with xr.open_dataset(rec.local_path) as input_data:
                time_var = self._get_variable_name("time")
                start_time = input_data[time_var].data.min()
                end_time = input_data[time_var].data.max()
                return TimeRange(start_time, end_time)
        raise ValueError(
            "A NetcdfProduct needs a local file to determine temporal coverage"
            " but the 'local_path' attribute of the provided file record "
            "does not point to an existing file."
        )

    def get_spatial_coverage(self, rec: FileRecord) -> Geometry:
        """
        Determine the spatial coverage of a data file.

        Args:
            rec: A file record representing the file of which to determine
                the spatial extent.

        Return:
            A 'Geometry' object representing the spatial that the given
            datafile covers.
        """
        if rec.local_path is not None and rec.local_path.exists():
            with xr.open_dataset(rec.local_path) as input_data:
                lons = input_data[self._get_variable_name("longitude")].data
                lats = input_data[self._get_variable_name("latitude")].data
                return geometry.parse_swath(lons, lats)

        raise ValueError(
            "A NetcdfProduct needs a local file to determine spatial coverage"
            " but the 'local_path' attribute of the provided file record "
            "does not point to an existing file."
        )

    def open(self, rec: FileRecord) -> xr.Dataset:
        """
        This function should read a given data file into an xarray.Dataset.

        Args:
            rec: A file record representing the file to be opened.

        Return:
            An 'xarray.Dataset' that contains the data of the provided file.
        """
        return xr.open_dataset(rec.local_path)



"""
pansat.geometry
===============

The pansat geometry provides functions to represent the spatial coverage
of data files using geometrical objects.
"""
import numpy as np
from shapely.geometry import Polygon, MultiPolygon
from shapely.validation import make_valid
from shapely.ops import unary_union


def parse_point(xml_point):
    """
    Parse point from XML file.
    """
    lon = float(xml_point[0].text)
    lat = float(xml_point[1].text)
    return [lon, lat]


def parse_polygon_xml(xml_polygon):
    """
    Parse polygon from XML file.
    """
    boundary = xml_polygon[0]
    points = np.array(list(map(parse_point, boundary)))
    dlons = points[:, 0] - points[0, 0]
    indices = np.where(np.abs(dlons) > 180)[0]
    for ind in indices:
        if points[ind, 0] > 0:
            points[ind, 0] -= 360
        else:
            points[ind, 0] += 360

    return make_valid(Polygon(points))


def parse_swath_xml(meta_data):
    """
    Parse shapes describing a satellite swath from metadata.

    Args:
        meta_data: XML tree containing the meta data.

    Return:
        List of polygons.
    """
    sdc = meta_data.find("SpatialDomainContainer")
    hsdc = sdc.find("HorizontalSpatialDomainContainer")

    polygons = list(map(parse_polygon_xml, hsdc))
    new_polygons = []
    poles = []

    for ind in range(len(polygons)):
        poly = polygons[ind]
        points = np.array(poly.convex_hull.exterior.coords)
        if any(points[:, 1] > 85):
            pole = Polygon([[-180, 70], [180, 70], [180, 90], [-180, 90]])
            poles.append(pole)
        elif any(points[:, 1] < -85):
            pole = Polygon([[-180, -70], [180, -70], [180, -90], [-180, -90]])
            poles.append(pole)
        else:
            new_polygons.append(poly)

    multi = unary_union(new_polygons + poles)
    return multi


def handle_poles(polygons):
    """
    Handle poles in a sequence of Polygons representing a swath from
    a polar orbiting satellite.

    The 2D representation of satellite observations breaks down along
    the poles. This function checks whether the observations extend
    outside of +/- 70 degree latitude and if so simply add a polygon
    covering all of the polar regions.

    Args:
        polygons: A list of shapely polygons.

    Return:
        The list of polygons with offending Polygons replaced.
    """
    for ind in range(len(polygons)):
        poly = polygons[ind]
        points = np.stack(poly.exterior.coords.xy, -1)
        if any(points[:, 1] > 70):
            poly_2 = Polygon([[-180, 75], [180, 75], [180, 90], [-180, 90]])
            polygons[ind] = poly_2
        if any(points[:, 1] < -70):
            poly_2 = Polygon([[-180, -75], [180, -75], [180, -90], [-180, -90]])
            polygons[ind] = poly_2

    poly = polygons[0]
    for other in polygons[1:]:
        poly = poly.union(other)
    return poly


def parse_swath(lons, lats, m=10, n=1) -> MultiPolygon:
    """
    Parse a swath as a 'shapely.Geometry'.

    Args:
        lons: A 2D array holding the longitude coordinates of the given
            observations.
        lats: A 2D array hodling the latitude coordinates of the
            observations.

    Return:
        A shapely geometry representing the observations.
    """
    n_i = lons.shape[0]
    n_j = lons.shape[1]
    d_i = n_i // m
    d_j = n_j // n
    ind_i = 0

    polys = []

    while ind_i < lons.shape[0]:
        ind_j = 0
        while ind_j < lons.shape[1]:
            lon_0_0 = lons[ind_i, ind_j]
            lat_0_0 = lats[ind_i, ind_j]
            lon_0_1 = lons[ind_i, min(ind_j + d_j, n_j - 1)]
            lat_0_1 = lats[ind_i, min(ind_j + d_j, n_j - 1)]
            lon_1_1 = lons[min(ind_i + d_i, n_i - 1), min(ind_j + d_j, n_j - 1)]
            lat_1_1 = lats[min(ind_i + d_i, n_i - 1), min(ind_j + d_j, n_j - 1)]
            lon_1_0 = lons[min(ind_i + d_i, n_i - 1), ind_j]
            lat_1_0 = lats[min(ind_i + d_i, n_i - 1), ind_j]

            polys.append(
                Polygon(
                    [
                        [lon_0_0, lat_0_0],
                        [lon_0_1, lat_0_1],
                        [lon_1_1, lat_1_1],
                        [lon_1_0, lat_1_0],
                        [lon_0_0, lat_0_0],
                    ]
                )
            )
            ind_j += d_j
        ind_i += d_i
    poly = handle_poles(polys)
    return make_valid(poly)



"""
pansat.download.providers.cloudnet
==================================

A data provider to download Cloudnet data.
"""
from datetime import datetime
from pathlib import Path
import requests

from pansat.download.providers.discrete_provider import DiscreteProvider


FILE_URL = "https://cloudnet.fmi.fi/api/files"


class CloudnetProvider(DiscreteProvider):
    """
    Provider class to download data from the cloudnet API.
    """

    @classmethod
    def get_available_products(cls):
        return [
            "ground_based::Cloudnet::radar",
            "ground_based::Cloudnet::iwc",
            "ground_based::Cloudnet::classification",
        ]

    def get_files_by_day(self, year, day):
        """
        Return files available on a given day.

        Args:
            year: Integer specifying the year.
            day: Integer specfiying the day of the yaer.

        Return:
            A list containing the available files.
        """
        date = datetime.strptime(f"{year}{day:03}", "%Y%j")
        payload = {
            "product": self.product.product_name,
            "date": date.strftime("%Y-%m-%d"),
        }
        if self.product.location is not None:
            payload["site"] = self.product.location
        response = requests.get(FILE_URL, payload)
        response.raise_for_status()
        files = [res["downloadUrl"].split("/")[-1] for res in response.json()]
        return [filename for filename in files if self.product.matches(filename)]

    def download_file(self, filename, destination):
        """
        Download a file.

        Args:
            filename: The name of the file.
            destination: Path to the file to which to write the
                 downloaded data.
        """
        filename = Path(filename)
        date, site, *_ = filename.name.split("_")
        payload = {
            "product": self.product.product_name,
            "site": site,
            "date": f"{date[:4]}-{date[4:6]}-{date[6:]}",
        }
        files = requests.get(FILE_URL, payload).json()

        url = files[0]["downloadUrl"]
        response = requests.get(url)
        with open(destination, "wb") as output:
            output.write(response.content)



"""
pansat.download.providers.eumetsat
==================================

This module provides the ``EUMETSATProvider`` class, which implements a data
provider class for the `EUMETSAT data store <https://www.data.eumetsat.int/>`_.
"""
import base64
from datetime import datetime
from ftplib import FTP
from pathlib import Path
import shutil

import requests

from pansat.download.providers.data_provider import DataProvider
from pansat.download.accounts import get_identity
from pansat.exceptions import CommunicationError

_PRODUCTS = {
    "MSG_Seviri": "EO:EUM:DAT:MSG:HRSEVIRI",
    "MSG_Seviri_IO": "EO:EUM:DAT:MSG:HRSEVIRI-IODC",
    "MHS_L1B": "EO:EUM:DAT:METOP:MHSL1",
    "AVHRR_L1B": "EO:EUM:DAT:METOP:AVHRRL1",
}


def _retrieve_access_token(key, secret):
    """
    Retrieve an access token for the EUMETSAT data store. The access token
    is required to download files. It is time limited and must be generated
    for every session.

    Args:
        key: The EUMETSAT API key. It is stored as user name in the pansat
            credentials. It is obtained from the EUMETSAT data store web API.
        secret: The corresponding password.
    """
    b64 = base64.b64encode((key + ":" + secret).encode())
    header = "Authorization: Basic " + b64.decode()
    data = {"grant_type": "client_credentials"}
    url = "https://api.eumetsat.int/token"
    req = requests.post(
        url,
        auth=requests.auth.HTTPBasicAuth(key, secret),
        data={"grant_type": "client_credentials"},
        headers={"Content-Type": "application/x-www-form-urlencoded"},
    )
    if not req.ok:
        raise CommunicationError(
            "Retrieving the EUMETSAT access token failed with the following"
            f"error:\n {req.text}"
        )
    return req.json()


###############################################################################
# Access Token
###############################################################################


class AccessToken:
    """
    This class represents an AccessToken for the EUMETSAT data store.
    """

    def __init__(self, key, secret):
        """
        Obtain access token.

        Args:
            key: The EUMETSAT API key. It is stored as user name in the pansat
                credentials. It is obtained from the EUMETSAT data store web API.
            secret: The corresponding password.
        """
        token_data = _retrieve_access_token(key, secret)
        self.token = token_data["access_token"]
        self.lifetime = token_data["expires_in"]
        self.created = datetime.now()

    @property
    def valid(self):
        """
        Whether the key is still valid.
        """
        now = datetime.now()
        age = (now - self.created).total_seconds()
        return age < self.lifetime

    def __repr__(self):
        return f"AccessToken(created={self.created}, lifetime={self.lifetime})"

    def __str__(self):
        return str(self.token)

    def renew(self, key, secret):
        """
        Renew access token.
        """
        self.__init__(key, secret)

    def ensure_valid(self):
        if not self.valid:
            self.renew()


class Collection:
    """
    Helper dataset for parsing collections.
    """

    def __init__(self, identifier, url):
        """
        Retrieve info for collection.
        """
        self.identifier = identifier
        self.url = url

        req = requests.get(url)
        if not req.ok:
            raise CommunicationError(
                f"Retrieving the collection {identifier} from the EUMETSAT data "
                f"store failed with the following error:\n {req.text}."
            )
        data = req.json()["collection"]
        self.title = data["properties"]["title"]
        self.abstract = data["properties"]["abstract"]

        def __repr__(self):
            return f"Collection({self.collection_id}, {self.title})"


class EUMETSATProvider(DataProvider):
    """
    Base class for data products available from the EUMETSAT data store.
    """

    base_url = "https://api.eumetsat.int"

    @staticmethod
    def get_collections():
        """
        Return collections available from the Eumetsat data
        store.

        Return:
            A dict mapping names of the available collections to the
            URLs at which they are available.
        """
        # b64 = base64.b64encode((key + ":" + secret).encode())
        url = "https://api.eumetsat.int/data/browse/collections"
        req = requests.get(
            url,
            params={"format": "json"},
        )
        if not req.ok:
            raise CommunicationError(
                "Retrieving the collections from the EUMETSAT data store failed"
                f"with the following error:\n {req.text}"
            )
        data = req.json()
        collections = []
        for link in data["links"]:
            identifier = link["title"]
            url = link["href"]
            collections.append(Collection(identifier, url))

        return collections

    def __init__(self, product):
        """
        Create a new product instance.

        Args:

            product(``Product``): Product class object with specific product for ICARE
        """
        if str(product) not in _PRODUCTS:
            available_products = list(_PRODUCTS.keys())
            raise ValueError(
                f"The product {product} is  not a available from the Eumetsat data"
                f" provider. Currently available products are: "
                f"{available_products}."
            )
        super().__init__()
        self.product = product
        self.page_size = 100
        self._token = None

    @property
    def token(self):
        """The identification token for the API."""
        if self._token is None:
            user, password = get_identity("EUMETSAT")
            self._token = AccessToken(user, password)
        if not self._token.valid:
            user, password = get_identity("EUMETSAT")
            self._token.renew(user, password)
        return self._token

    def get_available_products():
        """List of the products available from this provider."""
        return _PRODUCTS.keys()

    def get_files_in_range(self, start, end, bounding_box=None):
        """
        Return files in available in range.

        Args:
            start: Start of the time range.
            end: End of the time range.
            bounding_box: An optional bounding box specifying a region
                of interest (ROI). If given, only files containing
                observations over this regions will be returned.

        Return:
             A list of URLs of available files.
        """
        product_id = _PRODUCTS[str(self.product)]

        parameters = {
            "format": "json",
            "pi": product_id,
            "si": 0,
            "c": self.page_size,
            "dtstart": start.strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
            "dtend": end.strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
        }
        if bounding_box is not None:
            bb_str = ",".join(map(str, bounding_box))
            parameters["bbox"] = bb_str
        url = self.base_url + f"/data/search-products/os"

        links = []

        # Helper function to extract links from response data.
        def get_link(feature):
            return feature["properties"]["links"]["data"][0]["href"]

        total_results = self.page_size
        start_index = 0
        while start_index < total_results:
            parameters["si"] = start_index
            with requests.get(url, params=parameters) as r:
                r.raise_for_status()

                datasets = r.json()
                features = datasets["features"]
                links += map(get_link, datasets["features"])
            start_index += self.page_size

        links = []

        # Helper function to extract links from response data.
        def get_link(feature):
            return feature["properties"]["links"]["data"][0]["href"]

        total_results = self.page_size
        start_index = 0
        while start_index < total_results:
            parameters["si"] = start_index
            with requests.get(url, params=parameters) as r:
                datasets = r.json()
                features = datasets["features"]
                links += map(get_link, datasets["features"])
            start_index += self.page_size

        return links

    def download_file(self, link, destination):
        """
        Download a specific file.

        Args:
            link: Link pointing towards the file to download.:
            destination: Where to store the file.

        Return:
            The path of the local file to which the file content
            was written.
        """
        destination = Path(destination)
        if not destination.exists():
            destination.mkdir(exist_ok=True, parents=True)

        self.token.ensure_valid()
        params = {"access_token": str(self.token)}

        with requests.get(link, stream=True, params=params) as r:
            if not r.ok:
                raise CommunicationError(
                    f"Downloading the file {link} from the EUMETSAT data "
                    f"store failed token failed with the following"
                    f"error:\n {r.text}."
                )
            filename = link.split("/")[-1]
            dest = destination / (filename + ".zip")
            with open(dest, "wb") as f:
                shutil.copyfileobj(r.raw, f)
        return dest

    def download(self, start, end, destination):
        """
        Download all products within a given time range.

        Args:
            start: Start time
            end: End time.
            destination: Path to where to store the downloaded files.

        Return:
            List of the local paths of all downloaded files.
        """
        destination = Path(destination)

        links = self.get_files_in_range(start, end)
        downloads = []

        for link in links:
            filename = self.download_file(link, destination)
            downloads.append(filename)

        return downloads



"""
pansat.products.satellite.gpm
=============================

This module defines the GPM product class, which is used to represent all
GPM products.
"""
import re
from datetime import datetime
from itertools import dropwhile
from pathlib import Path

import numpy as np
import pandas as pd

import pansat.download.providers as providers
from pansat.products.product import Product
from pansat.products.product_description import ProductDescription
from pansat.exceptions import NoAvailableProvider
from pansat.formats.hdf5 import HDF5File
from pansat import geometry


class GPMProduct(Product):
    """
    Base class representing GPM products.
    """

    def __init__(self, level, platform, sensor, name, version, variant, description):
        self.level = level
        self.platform = platform
        self.sensor = sensor
        self.name = name
        self.version = version
        self.variant = variant
        self._description = description

        if self.variant:
            variant = "-" + self.variant
        else:
            variant = ""
        self.filename_regexp = re.compile(
            rf"{self.level}{variant}\.{self.platform}\.{self.sensor}"
            rf"\.{self.name}([\w-]*).(\d{{8}})-"
            r"S(\d{6})-E(\d{6})\.(\w*)\.((\w*)\.)?(HDF5|h5|nc|nc4)"
        )

    @property
    def variables(self):
        return []

    @property
    def description(self):
        return self._description

    def matches(self, filename):
        """
        Determines whether a given filename matches the pattern used for
        the product.

        Args:
            filename(``str``): The filename

        Return:
            True if the filename matches the product, False otherwise.
        """
        return self.filename_regexp.match(filename)

    def filename_to_date(self, filename):
        """
        Extract timestamp from filename.

        Args:
            filename(``str``): Filename of a GPM product.

        Returns:
            ``datetime`` object representing the timestamp of the
            filename.
        """
        path = Path(filename)
        match = self.filename_regexp.match(path.name)

        # Some files of course have to follow a different convetion.
        if match is None:
            date_string = "20" + path.name.split("_")[2]
        else:
            date_string = match.group(2) + match.group(3)
        date = datetime.strptime(date_string, "%Y%m%d%H%M%S")
        return date

    def filename_to_start_time(self, filename):
        path = Path(filename)
        match = self.filename_regexp.match(path.name)
        date_string = match.group(2) + match.group(3)
        date = datetime.strptime(date_string, "%Y%m%d%H%M%S")
        return date

    def filename_to_end_time(self, filename):
        path = Path(filename)
        match = self.filename_regexp.match(path.name)
        date_string = match.group(2) + match.group(4)
        date = datetime.strptime(date_string, "%Y%m%d%H%M%S")
        return date

    def get_spatial_coverage(self, rec):
        """
        Create geometry representing the spatial coverage of a data file.

        Args:
            rec: A 'FileRecord' object pointing to a data file.

        Return:
            A geometry object representing the the spatial coverage.
        """
        if rec.local_path is not None and rec.local_path.exists():
            with HDF5File(str(rec.local_path), "r") as input_data:
                lats = input_data["S1/Latitude"][:]
                lons = input_data["S1/Longitude"][:]
                valid = np.where(np.any(lons >= -180, 0))[0]
                left = valid[0]
                right = valid[-1]
                return geometry.parse_swath(lons[:, left:right], lats[:, left:right])
        raise ValueError(
            "A NetcdfProduct needs a local file to determine temporal coverage"
            " but the 'local_path' attribute of the provided file record "
            "does not point to an existing file."
        )

    def _get_provider(self):
        """Find a provider that provides the product."""
        available_providers = [
            p
            for p in providers.ALL_PROVIDERS
            if str(self) in p.get_available_products()
        ]
        if not available_providers:
            raise NoAvailableProvider(
                f"Could not find a provider for the" f" product {str(self)}."
            )
        return available_providers[0]

    @property
    def default_destination(self):
        """
        The default destination for GPM product is
        ``GPM/<product_name>``>
        """
        return Path("GPM") / Path(self.name)

    def __str__(self):
        if self.variant:
            variant = f"-{self.variant}"
        else:
            variant = ""
        s = f"GPM_{self.level}{variant}_{self.platform}_{self.sensor}"
        return s

    def download(self, start_time, end_time, destination=None, provider=None):
        """
        Download data product for given time range.

        Args:
            start_time(``datetime``): ``datetime`` object defining the start
                 date of the time range.
            end_time(``datetime``): ``datetime`` object defining the end date
                 of the of the time range.
            destination(``str`` or ``pathlib.Path``): The destination where to
                 store the output data.
        """

        if not provider:
            provider = self._get_provider()

        if not destination:
            destination = self.default_destination
        else:
            destination = Path(destination)
        destination.mkdir(parents=True, exist_ok=True)
        provider = provider(self)

        return provider.download(start_time, end_time, destination)

    def open(self, filename):
        """
        Open file as xarray dataset.

        Args:
            filename(``pathlib.Path`` or ``str``): The GPM file to open.
        """
        from pansat.formats.hdf5 import HDF5File

        with HDF5File(filename, "r") as file_handle:
            return self.description.to_xarray_dataset(file_handle, globals())


def _extract_scantime(scantime_group):
    """
    Extract scan time as numpy object.

    This function is use

    Args:
         scantime_group: The HDF5 group containing the scantime data.

    Returns:
         numpy.datetime64 object representing the scantime.
    """
    years = scantime_group["Year"][:]
    months = scantime_group["Month"][:]
    days = scantime_group["DayOfMonth"][:]
    hours = scantime_group["Hour"][:]
    minutes = scantime_group["Minute"][:]
    seconds = scantime_group["Second"][:]
    milli_seconds = scantime_group["MilliSecond"][:]
    n_dates = years.size
    dates = np.zeros(n_dates, dtype="datetime64[ms]")
    for i in range(n_dates):
        year = years[i]
        month = months[i]
        day = days[i]
        hour = hours[i]
        minute = minutes[i]
        second = seconds[i]
        milli_second = milli_seconds[i]
        dates[i] = np.datetime64(
            f"{year:04}-{month:02}-{day:02}"
            f"T{hour:02}:{minute:02}:{second:02}"
            f".{milli_second:03}"
        )
    return dates


def _parse_products():
    module_path = Path(__file__).parent
    for filename in module_path.iterdir():
        if filename.match("*.ini") and filename.name != "gprof.ini":
            description = ProductDescription(filename)
            python_name = filename.name.split(".")[0]
            level = description.properties["level"]
            platform = description.properties["platform"]
            sensor = description.properties["sensor"]
            name = description.properties["name"]
            version = int(description.properties["version"])
            variant = description.properties["variant"]
            globals()[python_name] = GPMProduct(
                level, platform, sensor, name, version, variant, description
            )


_parse_products()

################################################################################
# GPROF products
################################################################################


class GPROFProduct(GPMProduct):
    """
    Specialization of GPM product for GPROF products, which all have the same
    data format.
    """

    def __init__(self, platform, sensor, version, variant=""):
        module_path = Path(__file__).parent
        description = ProductDescription(module_path / "gprof.ini")
        super().__init__("2A", platform, sensor, "GPROF", version, variant, description)


l2a_gprof_gpm_gmi = GPROFProduct("GPM", "GMI", 5)
l2a_gprof_metopb_mhs = GPROFProduct("METOPB", "MHS", 5)


################################################################################
# GPM merged IR
################################################################################


class GPMMergeIR:
    """
    The GPM merged IR product.
    """

    def __init__(self):
        pattern = r"merg_(\d{10,10})_4km-pixel.nc"
        self.filename_regexp = re.compile(pattern)

    def matches(self, filename):
        """
        Determines whether a given filename matches the pattern used for
        the product.

        Args:
            filename(``str``): The filename

        Return:
            True if the filename matches the product, False otherwise.
        """
        return self.filename_regexp.match(filename)

    def filename_to_date(self, filename):
        """
        Extract timestamp from filename.

        Args:
            filename(``str``): Filename of a GPM product.

        Returns:
            ``datetime`` object representing the timestamp of the
            filename.
        """
        path = Path(filename)
        match = self.filename_regexp.match(path.name)
        date_string = match.group(1)
        date = datetime.strptime(date_string, "%Y%m%d%H")
        self.variant = ""
        return date

    def _get_provider(self):
        """Find a provider that provides the product."""
        available_providers = [
            p
            for p in providers.ALL_PROVIDERS
            if str(self) in p.get_available_products()
        ]
        if not available_providers:
            raise NoAvailableProvider(
                f"Could not find a provider for the" f" product {str(self)}."
            )
        return available_providers[0]

    @property
    def default_destination(self):
        """
        The default destination for GPM product is
        ``GPM/<product_name>``>
        """
        return Path("GPM") / Path(self.name)

    @property
    def name(self):
        return "gpm_mergeir"

    def __str__(self):
        return self.name

    def download(self, start_time, end_time, destination=None, provider=None):
        """
        Download data product for given time range.

        Args:
            start_time(``datetime``): ``datetime`` object defining the start
                 date of the time range.
            end_time(``datetime``): ``datetime`` object defining the end date
                 of the of the time range.
            destination(``str`` or ``pathlib.Path``): The destination where to
                 store the output data.
        """

        if not provider:
            provider = self._get_provider()

        if not destination:
            destination = self.default_destination
        else:
            destination = Path(destination)
        destination.mkdir(parents=True, exist_ok=True)
        provider = provider(self)

        return provider.download(start_time, end_time, destination)

    def open(self, filename):
        """
        Open file as xarray dataset.

        Args:
            filename(``pathlib.Path`` or ``str``): The GPM file to open.
        """
        from pansat.formats.hdf5 import HDF5File

        file_handle = HDF5File(filename, "r")
        return self.description.to_xarray_dataset(file_handle, globals())


gpm_mergeir = GPMMergeIR()


def _imerg_parse_time(seconds_since_1970):
    """
    Helper function to convert time from IMERG HDF5 files to
    numpy datetime.
    """
    return np.datetime64("1970-01-01T00:00:00") + seconds_since_1970[:].astype(
        "timedelta64[s]"
    )


def _gpm_l1c_parse_time(scan_time_group):
    """
    Helper function to convert time from GPM L1C to
    numpy datetime.
    """
    year = scan_time_group["Year"][:]
    month = scan_time_group["Month"][:]
    day = scan_time_group["DayOfMonth"][:]
    hour = scan_time_group["Hour"][:]
    minute = scan_time_group["Minute"][:]
    second = scan_time_group["Second"][:]
    milli_second = scan_time_group["MilliSecond"][:]

    return pd.to_datetime(
        pd.DataFrame(
            {
                "year": year,
                "month": month,
                "day": day,
                "hour": hour,
                "minute": minute,
                "second": second,
            }
        )
    )
    return scan_time


def parse_l1c_header(dataset, file_handle):
    """
    Callback to parse header of GPM L1C file.

    Args:
        dataset: ``xarray.Dataset`` containing the data loaded from the file.
        file_handle: File handle to the ``h5py`` ``File`` object.
    """
    attrs = {}
    for attr in file_handle.attrs["FileHeader"].decode().split("\n"):
        try:
            key, value = attr.split("=")
            attrs[key] = value[:-1]
        # If we can't split it, it's not an attribute.
        except ValueError:
            pass
    dataset.attrs.update(attrs)


def parse_frequencies(field):
    """
    Callback to parse frequencies from GPM L1C files.

    Args:
        field: The h5py variable containing the brightness temperatures.
    """
    lines = field.attrs["LongName"].decode().split()
    freqs = []
    c = 1
    c_name = f"{c})"

    for w1, w2 in zip(lines[:-1], lines[1:]):
        if w1 == c_name:
            freqs.append(float(w2.split("+")[0]))
            c += 1
            c_name = f"{c})"
    return np.array(freqs)


def parse_offsets(field):
    """
    Callback to parse frequency-offsets from GPM L1C files.

    Args:
        field: The h5py variable containing the brightness temperatures.
    """
    lines = field.attrs["LongName"].decode().split()
    lines.append("x")
    lines.append("x")
    offs = []
    c = 1
    c_name = f"{c})"
    for w1, w2, w3, w4 in zip(lines[:-4], lines[1:-2], lines[2:-1], lines[3:]):
        if w1 == c_name:
            freq = w2.split("+")
            if len(freq) > 1:
                offs.append(float(freq[1][1:]))
            elif w3.startswith("+"):
                if w3 == "+/-":
                    offs.append(float(w4))
                elif w3.startswith("+/-"):
                    offs.append(float(w3[3:]))
                else:
                    offs.append(float(w3[1:]))
            else:
                offs.append(0.0)
            c += 1
            c_name = f"{c})"

    return np.array(offs)



"""
pansat.catalog
==============

The ``catalog`` module provides functionality to organize, parse and
 list local and remote files.
"""
from pathlib import Path

import numpy as np
import xarray as xr

from pansat.time import to_datetime64
from pansat.file_record import FileRecord


class Catalog:
    """
    A catalog to manage collections of files of a given product.
    """

    def __init__(self, path, product):
        self.path = Path(path)
        self.product = product
        self.files = np.array(self._find_files(self.path))

        times = list(map(self.product.filename_to_date, self.files))
        self.times = np.array(list(map(to_datetime64, times)))

        indices = np.argsort(self.times)
        self.files = self.files[indices]
        self.times = self.times[indices]

    def _find_files(self, folder):
        """
        Recursively search for files in folder that match the
        product regexp.

        Args:
            folder: The directory to search.

        Return:
            A list of the filename that match the regexp of the
            product.
        """
        folder = Path(folder)
        files = []
        for path in folder.iterdir():
            if path.is_dir():
                files += self._find_files(path)
            else:
                if self.product.filename_regexp.match(path.name):
                    files.append(path)
        return files

    def find_file_covering(self, time):
        """
        Find file covering a given date.
        """
        time = to_datetime64(time)
        start_times = list(map(self.product.filename_to_start_time, self.files))
        start_times = np.array(list(map(to_datetime64, start_times)))
        end_times = list(map(self.product.filename_to_end_time, self.files))
        end_times = np.array(list(map(to_datetime64, end_times)))
        candidates = np.where((start_times <= time) * (end_times >= time))[0]
        return self.files[candidates]

    def load(self, start_time, end_time, dimension="time", load_callback=None):
        """
        Load data from files within a given time range.

        Args:
            start_time: Start time of the interval for which to load the data.
            end_time: End time of the interval for which to load the data.

        Return:
            An ``xarray.Dataset`` containing the loaded data.
        """
        start_time = to_datetime64(start_time)
        end_time = to_datetime64(end_time)
        indices = np.where((self.times >= start_time) * (self.times <= end_time))[0]
        datasets = []

        if load_callback is None:
            load_callback = self.product.open

        for ind in indices:
            dataset = load_callback(self.files[ind])
            if dimension not in dataset.dims:
                dataset = dataset.expand_dims("time")
                if dimension == "time":
                    filename = self.files[ind]
                    time = to_datetime64(self.product.filename_to_date(filename))
                    dataset["time"] = [time]

            datasets.append(dataset)

        return xr.concat(datasets, dim=dimension)


def find_files(product: "pansat.products.Prodcut", path: Path):
    """
    Find files of a given product.

    Args:
        product: A pansat product representing the product to find.
        path: A 'pathlib.Path' object pointing to a folder containing
            local files.

    Return:
        A list of file records pointing to the found files.
    """
    files = []
    for file_path in sorted(list(Path(path).glob("**/*"))):
        if product.matches(file_path.name):
            files.append(FileRecord(product, file_path))
    return files



import datetime

from pansat.products.product import Product


class OperaProduct(Product):
    """"""

    def __init__(self, product_name, description):
        self.product_name = product_name
        self._description = description
        self.filename_regexp = re.compile(
            rf"OPERA_{self.product}_(\d*)_(\d{2})_(\d{3})_(\d{2})_(\d{2}).hdf"
        )

    @property
    def variables(self):
        return []

    @property
    def description(self):
        return self._description

    def matches(self, filename):
        """
        Determines whether a given filename matches the pattern used for
        the product.

        Args:
            filename(``str``): The filename

        Return:
            True if the filename matches the product, False otherwise.
        """
        return self.filename_regexp.match(filename)

    def filename_to_date(self, filename):
        """
        Extract timestamp from filename.

        Args:
            filename(``str``): Filename of a GPM product.

        Returns:
            ``datetime`` object representing the timestamp of the
            filename.
        """
        path = Path(filename)
        match = self.filename_regexp.match(path.name)
        date_string = match.group(1) + match.group(2) + match.group(3) + match.group(4)
        date = datetime.strptime(date_string, "%y%j%H%M%S")
        return date

    def _get_provider(self):
        """Find a provider that provides the product."""
        available_providers = [
            p
            for p in providers.ALL_PROVIDERS
            if str(self) in p.get_available_products()
        ]
        if not available_providers:
            raise NoAvailableProviderError(
                f"Could not find a provider for the" f" product {self.name}."
            )
        return available_providers[0]

    @property
    def default_destination(self):
        """
        The default destination for GPM product is
        ``GPM/<product_name>``>
        """
        return Path("Opera") / Path(self.product_name)

    def __str__(self):
        s = f"OPERA_{self.product_name}"
        return s

    def download(self, start_time, end_time, destination=None, provider=None):
        """
        Download data product for given time range.

        Args:
            start_time(``datetime``): ``datetime`` object defining the start
                 date of the time range.
            end_time(``datetime``): ``datetime`` object defining the end date
                 of the of the time range.
            destination(``str`` or ``pathlib.Path``): The destination where to
                 store the output data.
        """

        if not provider:
            provider = self._get_provider()

        if not destination:
            destination = self.default_destination
        else:
            destination = Path(destination)
        destination.mkdir(parents=True, exist_ok=True)
        provider = provider(self)

        return provider.download(start_time, end_time, destination)

    def open(self, filename):
        """
        Open file as xarray dataset.

        Args:
            filename(``pathlib.Path`` or ``str``): The GPM file to open.
        """
        from pansat.formats.hdf5 import HDF5File

        file_handle = HDF5File(filename, "r")
        return self.description.to_xarray_dataset(file_handle, globals())



"""
pansat.roi
==========

Functions to define and work with regions of interest (ROIs).
"""
import numpy as np


class ROI:
    def __init__(self, lon_ll, lat_ll, lon_ur, lat_ur):
        self._components = np.array([lon_ll, lat_ll, lon_ur, lat_ur])

    @property
    def lon_min(self):
        return self._components[0]

    @property
    def lon_ll(self):
        return self._components[0]

    @property
    def lat_min(self):
        return self._components[1]

    @property
    def lat_ll(self):
        return self._components[1]

    @property
    def lon_max(self):
        return self._components[2]

    @property
    def lon_ur(self):
        return self._components[2]

    @property
    def lat_max(self):
        return self._components[3]

    @property
    def lat_ur(self):
        return self._components[3]

    def __iter__(self):
        yield from self._components

    def __repr__(self):
        return f"ROI({self.lon_ll}, {self.lat_ll}, {self.lon_ur}, {self.lat_ur})"

    def __getitem__(self, i):
        if (i < 0) or (i < 0):
            raise ValueError("The bounding box has only for elements.")
        return self._components[i]

    def __setitem__(self, i, val):
        if (i < 0) or (i < 0):
            raise ValueError("The bounding box has only for elements.")
        self._components[i] = val

    def to_geometry(self):
        """
        Return representation of ROI as ``shapely.geometry.Polygon``.
        """
        from shapely.geometry import Polygon

        return Polygon(
            [
                [self.lon_min, self.lat_min],
                [self.lon_max, self.lat_min],
                [self.lon_max, self.lat_max],
                [self.lon_min, self.lat_max],
            ]
        )


class PolygonROI:
    def __init__(self, points):
        self.points = points

    @property
    def lon_min(self):
        return self.points[:, 0].min()

    @property
    def lon_ll(self):
        return self.points[:, 0].min()

    @property
    def lat_min(self):
        return self.points[:, 1].min()

    @property
    def lat_ll(self):
        return self.points[:, 1].min()

    @property
    def lon_max(self):
        return self.points[:, 0].max()

    @property
    def lon_ur(self):
        return self.points[:, 0].max()

    @property
    def lat_max(self):
        return self.points[:, 1].max()

    @property
    def lat_ur(self):
        return self.points[:, 1].max()

    def __iter__(self):
        yield from self.points

    def __repr__(self):
        return f"PolyROI({self.points})"

    def to_geometry(self):
        """
        Return representation of ROI as ``shapely.geometry.Polygon``.
        """
        from shapely.geometry import Polygon

        return Polygon(self.points)


def _get_lats_and_lons(data):
    """
    Get latitude and longitude coordinates from xarray dataset.

    Searches for variables named 'lat' or 'latitude' and 'lon' and
    'longitude' and returns them as a pair.

    Args:
        data: An xarray dataset with satellite swath data.

    Return:
        A tuple ``(lat, lon)`` containing the latitude and longitude
        arrays of the dataset.
    """
    if "latitude" in data:
        lats = data.latitude
    elif "lat" in data:
        lats = data.lat
    else:
        raise ValueError("Neither 'lat' nor 'latitude' variable is present in dataset.")

    if "longitude" in data:
        lons = data.longitude
    elif "lon" in data:
        lons = data.lon
    else:
        raise ValueError(
            "Neither 'lon' nor 'longitude' variable is present in dataset."
        )

    return lats, lons


def find_overpasses(roi, data):
    lats, lons = _get_lats_and_lons(data)
    lats = lats.data
    lons = lons.data

    axes = tuple(range(lats.ndim))[1:]
    indices = np.where(
        np.any(
            (lats >= roi.lat_min)
            * (lats <= roi.lat_max)
            * (lons >= roi.lon_min)
            * (lons <= roi.lon_max),
            axis=axes,
        )
    )[0]

    if len(indices) == 0:
        return []

    changes = np.diff(indices)
    gaps = np.where(changes > 1)[0]

    first_dim = next(iter(data.dims.keys()))
    i_start = indices[0]

    slices = []

    offset = 0
    for g in gaps:
        i_start = indices[0]
        i_end = indices[g - offset]
        indices = indices[g - offset + 1 :]
        offset = g + 1
        slices.append(data[{first_dim: slice(i_start, i_end + 1)}])

    i_start = indices[0]
    i_end = indices[-1]
    slices.append(data[{first_dim: slice(i_start, i_end + 1)}])
    return slices


def any_inside(roi, data):
    lats, lons = _get_lats_and_lons(data)
    lats = lats.data
    lons = lons.data

    axes = tuple(range(lats.ndim))[1:]
    return np.any(
        np.any(
            (lats >= roi.lat_min)
            * (lats <= roi.lat_max)
            * (lons >= roi.lon_min)
            * (lons <= roi.lon_max),
            axis=axes,
        )
    )


def some_inside(roi, data, fraction=0.5):
    lats, lons = _get_lats_and_lons(data)
    lats = lats.data
    lons = lons.data

    axes = tuple(range(lats.ndim))[1:]
    inside = np.mean(
        (lats >= roi.lat_min)
        * (lats <= roi.lat_max)
        * (lons >= roi.lon_min)
        * (lons <= roi.lon_max),
        axis=axes,
    )
    return np.any(inside > fraction)
